#!/bin/env python
# -*- coding: utf-8 -*-
# -*- python -*-
# Author: Luo Jinghua

import urllib
import httplib
import htmlentitydefs
import time
import re
import os
import StringIO
import copy
import sys
import time

from xml.dom import minidom
import BeautifulSoup

KANPPS = 'http://kan.pps.tv'
MOVIE_LIST_PATH = '/movie_index.html'
MOVIE_LIST_URL = KANPPS + MOVIE_LIST_PATH

class PPSClass:
    def __init__ (self):
        self.id = 0
        self.title = ''
        self.url = ''
        self.pages = {}
        self.max_page = 1
        self.parent = None

    def __str__(self):
        return 'PPSClass<%d %s %s>' % (self.id, self.title, self.url)

    def __repr__(self):
        return repr(self.dictionary())

    def dictionary(self):
        d = {}
        d['id'] = self.id
        d['url'] = self.url
        d['title'] = self.title
        return d

    def parseid (self, url):
        return int(re.sub('.*/(\d+)/.*', r'\1', url))

    def parse (self, node):
        self.title = unicode(node.contents[0]).encode('utf-8')
        self.url = KANPPS + node['href'].encode('utf-8')
        self.id = self.parseid(self.url)
        return self

    def load(self, d):
        self.__dict__.update(d)

    def resetMovies(self):
        self.pages = {}
        self.max_page = 1

    def addMovies(self, movies, page_id = 0):
        movies = copy.copy(movies)
        for movie in movies:
            movie.parent = self

        if not self.pages.has_key(page_id):
            self.pages[page_id] = []
        self.pages[page_id] += movies

    def setMovies(self, movies, page_id = 0, clone = True):
        if clone:
            movies = copy.copy(movies)
        for movie in movies:
            movie.parent = self

        self.pages[page_id] = movies

    def getMovies(self):
        pages = self.getPages()
        movies = []
        for page in pages:
            movies += page
        return movies

    def getPage(self, page_id):
        assert (page_id >= 0 and page_id < self.max_page)
        if self.pages.has_key(page_id):
            return self.pages[page_id]
        return []

    def getPages(self):
        keys = self.pages.keys()
        keys.sort()
        pages = []
        for key in keys:
            pages.append(self.pages[key])
        return pages

    def getMaxPage(self):
        return self.max_page

class PPSFile:
    elements = [ 'id', 'ci', 'size', 'url' ]

    def __str__ (self):
        return 'PPSFile<%d %s %s>' % (self.id, self.title, self.url)

    def __repr__(self):
        return repr(self.dictionary())

    def __eq__ (self, other):
        for attr in PPSFile.elements:
            if getattr (self, attr) != getattr (other, attr):
                #print attr, getattr (self, attr), getattr (other, attr)
                return False
        return True

    def dictionary(self):
        d = {}
        d['id'] = self.id
        d['url'] = self.url
        d['title'] = self.title
        return d

    def __init__ (self):
        self.id = 0
        self.ci = 0
        self.size = 0
        self.title = ''
        self.url = ''
        self.parent = None

    def parse (self, s, ci):
        idendpos = s.find(']')
        valuestartpos = s.find('=') + 2
        valueendpos = len(s) - 1
        values = s[valuestartpos:valueendpos].split('|||')
        self.id = int(s[6:idendpos])
        self.title = values[0]
        self.url = values[2]
        self.ci = ci
        return self

    def load(self, d):
        self.__dict__.update(d)

class PPSMovie:
    elements = [ 'id', 'title', 'director', 'actor',
                 'area', 'size', 'pubtime', 'length',
                 'lang', 'score', 'desc', 'image',
                 'cn', 'playerurl']
    def __str__(self):
        return 'PPSMovie<%d %s %s %d %s>' % (self.id, self.title,
                                             self.score, self.cn,
                                             self.actor)

    def __eq__ (self, other):
        for i in PPSMovie.elements:
            if getattr (self, i) != getattr (other, i):
                return False
        return self.files == other.files

    def __repr__(self):
        return repr(self.dictionary())

    def dictionary(self):
        d = {}
        for key in PPSMovie.elements:
            d[key] = getattr(self, key)
        return d

    def __init__ (self):
        self.id = 0
        self.title = ''
        self.director = ''
        self.actor = ''
        self.area = ''
        self.size = 0
        self.pubtime = ''
        self.length = ''
        self.lang = ''
        self.score = ''
        self.desc = ''
        self.image = ''
        self.cn = 1
        self.playerurl = ''
        self.baseurl = ''
        self.max_page = 1
        self.files = {}
        self.pixbuf = None
        self.parent = None

    def parseid (self, url):
        s = url[url.rfind('/') + 1:]
        if s.find('_') < 0:
            ids = s.split('.')
        else:
            ids = s.split('_')
        return int(ids[0])

    def parse_actors(self, node):
        actors = node.findAll('a')
        result = []
        for actor in actors:
            if actor.contents:
                result.append(extractString(actor.contents[-1]))
        result = {}.fromkeys(result).keys()
        return result

    def parse_area (self, node):
        area = unicode(extractString(node.contents[-1]), 'utf-8')
        if area.find (u'：') >= 0:
            area = area[area.find(u'：') + 1:]
        return area.encode('utf-8')

    def parse (self, node):
        self.files = {}
        self.pixbuf = None
        self.image = extractString(node.find('img')['src'])
        tr = node.find('dl', { "class" : "tr" })
        dt = node.find('dt')
        self.score = extractNumberString(dt.find('span').contents[0])
        self.title = extractString(dt.find('a').contents[0])
        self.cn = extractNumber(dt.contents[-1], 1)
        self.playerurl = KANPPS + extractString(dt.find('a')['href'])
        li = node.findAll('li')
        self.area = self.parse_area(li[0].find('span'))
        self.pubtime = extractString(li[0].contents[-1])
        self.actor = ', '.join(self.parse_actors(li[1]))
        self.desc = extractString(li[2].contents[-1])
        self.id = self.parseid(self.playerurl)
        self.smallimage = self.image
        return self

    def load(self, d):
        self.__dict__.update(d)
        self.smallimage = self.image

    def getFiles(self, page_id = 0):
        if page_id in self.files:
            return self.files[page_id]
        return []

    def getAllFiles(self):
        keys = self.files.keys()
        keys.sort()
        result = []
        for key in keys:
            result += self.files[key]
        return result

    def setFiles(self, files, page_id = 0):
        self.files[page_id] = copy.copy(files)
        for f in self.files[page_id]:
            f.parent = self

    def setMaxPage(self, max_page):
        self.max_page = max_page

    def getMaxPage(self):
        return self.max_page

def download(url, max_retry = 3, interval = 5):
    if not url:
        return ''
    for i in range (max_retry):
        try:
            res = urllib.urlopen (url).read ()
        except Exception, e:
            print "Coudn't open url: ", e
            res = None
            time.sleep (interval)
        if res:
            break
    return res

def gbk2utf8(s):
    return unicode(s, 'gb18030', 'ignore').encode('utf-8')

def extractString(s):
    return unicode(s).encode('utf-8')

def extractStrings(ss):
    result = []
    for s in ss:
        result.append(extractString(s))
    return result

def extractNumberString(s, default = ''):
    s = extractString(s)
    n = re.search(r'[-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?', s)
    if n:
        return n.group()
    return default

def extractNumber(s, default = 0):
    s = extractString(s)
    n = re.search(r'[-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?', s)
    if n:
        if '.' in n.group() or 'e' in n.group() or 'E' in n.group():
            return float(n.group())
        return int(n.group())
    return default

def extractNavigableStrings(node):
    result = []
    for content in node.contents:
        if type(content) is BeautifulSoup.NavigableString:
            result.append(extractString(content))
    return result

def parseMovieClassList(res):
    ppslist = gbk2utf8(res)
    client_list = '<dt>客户端列表</dt>'
    startpos = ppslist.find(client_list)
    if startpos < 0:
        return []
    endpos = ppslist[startpos:].find('</dd>')
    if endpos < 0:
        return []
    soup = BeautifulSoup.BeautifulSoup(ppslist[startpos:startpos + endpos + 5])
    result = []
    for l in soup.findAll('li'):
        cls = PPSClass()
        result.append(cls.parse(l.next))
    return result

def getMovieClassList():
    movie_index_fname = 'movie_index.html'
    if os.path.exists(movie_index_fname):
        ppslist = file(movie_index_fname).read()
    else:
        ppslist = download(MOVIE_LIST_URL)
        file(movie_index_fname, 'wb').write(ppslist)

    return parseMovieClassList(ppslist)

def parseMovieList(movie_list, s):
    s = gbk2utf8(s)
    soup = BeautifulSoup.BeautifulSoup(s)
    result = []
    for l in soup.findAll('div', { "class" : "pltr" }):
        movie = PPSMovie()
        result.append(movie.parse(l))
    return result

def parseMovieListPage(movie_list, s):
    tag = '<div class="pageNav">'
    tagpos = s.find(tag)
    if tagpos < 0:
        return (1, 1)
    startpos = s[tagpos:].find('<span>')
    endpos = s[tagpos:].find('</span>')
    pagestr = s[tagpos + startpos:tagpos + endpos]
    m = re.search('(\d+)/(\d+)', pagestr)
    result = m.groups()
    return (int(result[0]), int(result[1]))

def getMovieList(movie_list, retrieve_all = False):
    s = download(movie_list.url)
    page = parseMovieListPage(movie_list, s)
    result = parseMovieList(movie_list, s)
    if retrieve_all:
        url = movie_list.url
        pos = url.rfind('/')
        baseurl = url[:pos + 1]
        pos = url.rfind('.')
        suffix = url[pos:]
        for i in range(page[1] - 1):
            print 'retrieving %d of %d' % (i + 2, page[1])
            url = '%s%d%s' % (baseurl, i + 2, suffix)
            s = download(url)
            print url
            result += parseMovieList(movie_list, s)
    return result

def parseMovieFileList(movie, s):
    s = gbk2utf8(s)
    result = []
    ci = 0
    for f in re.findall('plist\[\d+\][^;]*', s):
        ppsfile = PPSFile()
        result.append(ppsfile.parse(f, ci))
        ci += 1
    return result

def getMovieFileListFrameUrl(s):
    s = gbk2utf8(s)
    soup = BeautifulSoup.BeautifulSoup(s)
    result = []
    play_iframe = soup.findAll('iframe', { 'id': 'play_iframe' })
    if not play_iframe:
        return ''
    return play_iframe[0]['src']

def getMovieFileListFrameBaseUrl(s):
    url = getMovieFileListFrameUrl(s)
    if not url:
        return url
    pos = url.rfind('/')
    if pos > 0:
        return url[0:pos]
    return url

def parseMoviePlayList(s, sort_list = False):
    s = gbk2utf8(s)
    soup = BeautifulSoup.BeautifulSoup(s)
    seein = soup.findAll('div', { 'class': 'onseeIn' })
    if not seein:
        return ()
    uls = seein[0].findAll('ul')
    if not uls:
        return ()
    items = uls[0].findAll('li')
    if not items:
        return ()

    files = []
    for item in items:
        f = PPSFile()
        fid = extractNumber(item['id'])
        refs = item.findAll('a')
        title = refs[0].contents[-1]
        url = refs[1]['href']
        f.id = fid
        f.title = unicode(title).encode('utf-8')
        f.url = unicode(url).encode('utf-8')
        files.append(f)

        if sort_list:
            def ppsfile_sort_key(f):
                return f.title
            files.sort(key = ppsfile_sort_key)

    pagenav = soup.findAll('div', { 'class': 'pageNav' })
    if not pagenav:
        return ()
    refs = pagenav[0].findAll('a')
    maxpage = int(refs[-1].contents[-1])
    spans = pagenav[0].findAll('span')
    curpage = int(spans[0].contents[-1])
    return (files, max(maxpage, curpage), curpage - 1)

def getMovieFileList(movie):
    s = download(movie.playerurl)
    return parseMovieFileList(movie, s)

class PPSList:
    def __init__(self):
        self.classes = []

    def reset(self):
        self.classes = []

    def getClasses(self):
        return self.classes

    def fetchClasses(self):
        s = download(MOVIE_LIST_URL)
        return s

    def parseClasses(self, s):
        if s:
            result = parseMovieClassList(s)
        else:
            result = []
        return result

    def updateClasses(self, classes):
        self.classes += copy.copy(classes)

    def fetchMovieList(self, movie_class, page_id = 0):
        url = movie_class.url
        if page_id != 0:
            assert (page_id < movie_class.max_page)
            pos = url.rfind('/')
            baseurl = url[:pos + 1]
            pos = url.rfind('.')
            suffix = url[pos:]
            url = '%s%d%s' % (baseurl, page_id + 2, suffix)
        s = download(url)
        return s

    def searchMovieList(self, keyword, page_id = 0):
        baseurl = 'http://kan.pps.tv/search/'
        keyword = unicode(keyword, 'utf-8').encode('gb18030')
        url = baseurl + urllib.quote(keyword) + '/'
        if page_id > 0:
            url += "%d.html" % (page_id + 1)
        s = download(url, interval = 10)
        #print keyword, url, gbk2utf8(s)
        return s

    def parseMovieList(self, movie_class, s, page_id = 0):
        page = parseMovieListPage(movie_class, s)
        result = parseMovieList(movie_class, s)
        for movie in result:
            movie.page_id = page_id
        if page_id == 0:
            movie_class.max_page = page[1]
        return result

    def updateMovieList(self, movie_class, movie_list, page_id = 0):
        if page_id == 0:
            movie_class.movies = []
        movie_class.movies += movie_list
        movie_class.pages[page_id] = copy.copy(movie_list)

    def fetchMovie(self, cls, movie):
        s = download(movie.playerurl)
        return s

    def parseMovie(self, cls, movie, s):
        result = parseMovieFileList(movie, s)
        return result

    def parseMoviePlayListUrl(self, cls, movie, s):
        result = getMovieFileListFrameBaseUrl(s)
        movie.baseurl = result
        return result

    def fetchMoviePlayList(self, cls, movie, page_id = 0):
        if not movie.baseurl:
            return None
        s = download(movie.baseurl + '/%d.html' % page_id)
        return s

    def parseMoviePlayList(self, cls, movie, s, page_id = 0):
        result = parseMoviePlayList(s, True)
        if result:
            max_page = result[1]
            if max_page > movie.getMaxPage():
                movie.setMaxPage(max_page)
        return result

    def updateMovie(self, cls, movie, file_list, page_id = 0):
        movie.setFiles(file_list, page_id)

if __name__ == '__main__':
    ## movie_class_list = getMovieClassList()

    ## for movie_class in movie_class_list:
    ##     print movie_class

    ## movie_list = getMovieList(movie_class_list[1])

    ## for movie in movie_list:
    ##     print movie

    ## file_list = getMovieFileList(movie_list[0])

    ## for f in file_list:
    ##     print f

    def test_get_file_list():
        s = download('http://kan.pps.tv/play/281874.html')
        file_list = parseMovieFileList(PPSMovie(), s)
        for f in file_list:
            print f

        sys.exit(0)

    def test_get_file_list_by_iframe():
        url = 'http://kan.pps.tv/play/298356.html'
        s = download(url)
        movie = PPSMovie()
        file_list = parseMovieFileList(movie, s)
        for f in file_list:
            print f

        if file_list:
            return
        baseurl = getMovieFileListFrameBaseUrl(s)
        print 'baseurl = ', baseurl

        first = baseurl + '/0.html'
        s = download(first)
        if not s:
            return

        last_maxpage = 0
        cur_page = 0
        while True:
            url = baseurl + '/%d.html' % cur_page
            s = download(url)
            if not s:
                break
            result = parseMoviePlayList(s)
            if not result:
                break
            files, max_page = result
            print 'maxpage = ', max_page, cur_page
            for f in files:
                print f

            if max_page == cur_page:
                break
            cur_page += 1

    def test_search ():
        ppslist = PPSList()
        s = ppslist.searchMovieList('周星驰')
        movie_list = ppslist.parseMovieList(PPSClass(), s)
        for m in movie_list:
            print m
        sys.exit(0)

    def test_ppslist():
        ppslist = PPSList()
        s = ppslist.fetchClasses()
        ppslist.updateClasses(ppslist.parseClasses(s))
        classes =  ppslist.getClasses()
        for cls in classes:
            print cls

        cls = classes[0]
        s = ppslist.fetchMovieList(cls)
        ppslist.updateMovieList(cls, ppslist.parseMovieList(cls, s))
        page_id = 1
        while page_id < cls.max_page:
            s = ppslist.fetchMovieList(cls, page_id)
            ppslist.updateMovieList(cls, ppslist.parseMovieList(cls, s, page_id),
                                    page_id)
            page_id += 1

        movies = cls.getMovies()
        for movie in movies:
            print movie

        page = cls.getPage(0)
        for movie in page:
            s = ppslist.fetchMovie(cls, movie)
            ppslist.updateMovie(cls, movie, ppslist.parseMovie(cls, movie, s))

        for movie in page:
            print movie
            for f in movie.files:
                print f
            print

        movie = page[0]
        ppsfile = movie.files[0]
        d = ppsfile.dictionary()

        print ppsfile, d
        ppsfile1 = PPSFile()
        ppsfile1.load(d)
        print ppsfile1, ppsfile == ppsfile1

    #test_ppslist()
    #test_search()
    #test_get_file_list()
    #test_get_file_list_by_iframe()
