import totem
import gobject, gtk, gconf
gobject.threads_init()
import urllib
import httplib
import htmlentitydefs
import threading
import time
import re
import os
import random
import gettext
import copy
import pickle

from xml.dom import minidom
from traceback import print_exc

from ppstream_translation import _

import ppslist
import PPSVideoList

try:
	from hashlib import md5

	def url_digest (url):
		m = md5 ()
		m.update (url)
		return m.hexdigest ()
except ImportError:
	import md5

	def url_digest (url):
		m = md5.new ()
		m.update (url)
		return m.hexdigest ()

def unescape_charref (ref):
    name = ref[2:-1]
    base = 10
    if name[0] == 'x':
        name = name[1:]
        base = 16
    return unichr (int(name, base))

def replace_entities (match):
    ent = match.group ()
    if ent[1] == "#":
        return unescape_charref (ent)

    repl = htmlentitydefs.name2codepoint.get (ent[1:-1])
    if repl is not None:
        repl = unichr (repl)
    else:
        repl = ent
    return repl

def unescape (data):
    return re.sub (r"&#?[A-Za-z0-9]+?;", replace_entities, data)

def unescape_xml (data):
	ustr = unicode(data, 'utf-8')
	return unescape(ustr).encode ('utf-8')

gconf_key = '/apps/totem/plugins/ppstream'
pps_cache_dir = os.path.expanduser ('~/.local/share/totem/plugin/ppstream')


pps_xmls_dir = os.path.join (pps_cache_dir, "xmls")
pps_images_dir = os.path.join (pps_cache_dir, "images")
for p in [ pps_xmls_dir, pps_images_dir ]:
	if not os.path.exists (p):
		os.makedirs (p)

### special category ids
CATEGORY_ID_FAVORITES = 0
CATEGORY_ID_RECENT = -1
CATEGORY_ID_SEARCH = -2

### special category index
CATEGORY_FAVORITES = 0
CATEGORY_RECENT = 1
CATEGORY_SEARCH = 2
CATEGORY_NUM_SPECIAL = 3

### maximium number of recents
MAX_RECENT = 25

### maximium cached time
CATEGORY_MAX_CACHED_TIME = 2 * 24 * 60 * 60
MOVIES_MAX_CACHED_TIME = 7 * 24 * 60 * 60
MOVIE_MAX_CACHED_TIME = 1 * 24 * 60 * 60

def xml_node_value (node):
	return node.firstChild.nodeValue

def xml_child_value (node, id):
	child = node.getElementsByTagName(id)[0].firstChild;
	if child:
		return child.nodeValue
	return None

class ProcessThread (threading.Thread):
	def __init__ (self, data, process, callback, *args, **kwargs):
		self.data = data
		self.process = process
		self.callback = callback
		self.args = args
		self.kwargs = kwargs
		threading.Thread.__init__ (self)

	def run (self):
		try:
			res = self.process (self.data, *self.args, **self.kwargs)
		except Exception, e:
			print "Couldn't process data: ", e
			print_exc()
			res = None

		gobject.idle_add (self.publish_results, res)

	def publish_results(self, res):
		self.callback (res, *self.args, **self.kwargs)
		return False

class DownloadThread (threading.Thread):
	def __init__ (self, url, callback, *args, **kwargs):
		self.url = url
		self.callback = callback
		self.args = args
		self.kwargs = kwargs
		threading.Thread.__init__ (self)

	def run (self):
		for i in range (3):
			try:
				res = urllib.urlopen (self.url).read ()
			except Exception, e:
				print "Coudn't open url: ", e
				res = None
			if res:
				break
		gobject.idle_add (self.publish_results, res)

	def publish_results(self, res):
		self.callback (res, *self.args, **self.kwargs)
		return False

class RetrieveImageTask:
	pass

class RetrieveImage (threading.Thread):
	def __init__ (self):
		self.queue = []
		self._cache = {}
		self._lock = threading.Lock ()
		self._done = False
		threading.Thread.__init__ (self)

	def retrieve (self, url, callback, *args):
		task = RetrieveImageTask ()
		task.url = url
		task.callback = callback
		task.args = args
		self._lock.acquire ()
		self.queue.append (task)
		self._lock.release ()

	def cancel_tasks (self):
		self._lock.acquire (True)
		self.queue = []
		self._cache = {}
		self._lock.release ()

	def cancel (self):
		self._done = True
		self._lock.acquire (True)
		self.queue = []
		self._cache = {}
		self._lock.release ()

	def is_busy (self):
		self._lock.acquire (True)
		busy = len (self.queue) != 0
		self._lock.release ()

		return busy

	def get_remaning_tasks (self):
		self._lock.acquire (True)
		num = len (self.queue)
		self._lock.release ()

		return num

	def image_name_from_url (self, url):
		ext = url.rfind('.')
		if ext >= 0:
			ext = url[ext:]
		else:
			ext = ''
		filename = url_digest (url) + ext
		return os.path.join (pps_cache_dir, "images", filename)

	def load_image_from_disk (self, url):
		filename = self.image_name_from_url (url)
		if not os.access (filename, os.R_OK):
			return None

		#print 'loading image', filename
		try:
			pixbuf = gtk.gdk.pixbuf_new_from_file (filename)
		except gobject.GError:
			return None
		return pixbuf

	def store_image_to_disk (self, url, origin):
		filename = self.image_name_from_url (url)
		df = None
		sf = None
		try:
			df = file (filename, 'wb')
			sf = file (origin, 'rb')
			while True:
				buf = sf.read (4096)
				if not buf:
					break
				df.write (buf)
		except Exception, e:
			print 'Failed to copy ', origin, ' to ', filename
			print e
		finally:
			if df:
				df.close ()
			if sf:
				sf.close ()

	def load_pixbuf (self, url):
		pixbuf = self.load_image_from_disk (url)
		if pixbuf:
			return pixbuf

		try:
			filename, headers = urllib.urlretrieve (url)
		except Exception, e:
			print "Couldn't download ", url
			print "Error: ", e
			return None

		try:
			pixbuf = gtk.gdk.pixbuf_new_from_file (filename)
		except gobject.GError:
			return None

		self.store_image_to_disk (url, filename)
		os.unlink (filename)
		return pixbuf

	def load_from_cache (self, url):
		pixbuf = None
		self._lock.acquire (True)
		if url in self._cache:
			pixbuf = self._cache [url]
		self._lock.release ()
		return pixbuf

	def put_in_cache (self, url, pixbuf):
		self._lock.acquire (True)
		self._cache[url] = pixbuf
		self._lock.release ()

	def run (self):
		import time
		while not self.done:
			task = None
			self._lock.acquire (True)
			if len (self.queue):
				#print 'pending tasks:', len(self.queue)
				task = self.queue.pop (0)
			self._lock.release ()

			if not task:
				time.sleep (0.1)
				continue

			pixbuf = self.load_from_cache (task.url)
			if not pixbuf:
				#print 'loading:', task.url
				pixbuf = self.load_pixbuf (task.url)
				if pixbuf:
					self.put_in_cache (task.url, pixbuf)
					time.sleep (0.10)

			if not pixbuf:
				continue

			#print 'image loaded:', task.url
			gobject.timeout_add (50, self.publish_result, pixbuf, task)

	def publish_result (self, pixbuf, task):
		task.callback (pixbuf, *task.args)
		return False

	@property
	def done (self):
		self._lock.acquire (True)
		res = self._done
		self._lock.release ()
		return res

class PPStream (totem.Plugin):
	def __init__ (self):
		totem.Plugin.__init__ (self)
		self.debug = False
		self.gstreamer_plugins_present = True

		self.gconf = gconf.client_get_default ()

		self.max_results = 20
		self.button_down = False

		self.start_index = {}
		self.entry = {}

		self.current_treeview_name = ""
		self.notebook_pages = []

		self.vadjust = {}
		self.liststore = {}
		self.treeview = {}

		self.subclass_index = 0
		self.ppslist = ppslist.PPSList()

		self.current_movie = None
		self.movie_class_path = ()

        def load_ui (self, filename, fatal, parent, user_data):
		datadir = os.path.dirname (__file__)
                filename = os.path.join (datadir, filename)
                builder = gtk.Builder ()
                builder.set_translation_domain ('totem-pps')
                builder.add_from_file (filename)
                builder.connect_signals (self, user_data)
                return builder

	def activate (self, totem_object):
		"""Check for the availability of the ppssrc GStreamer plugins"""
		bvw_name = totem_object.get_video_widget_backend_name ()

		self.can_recode_mrl = False
		if bvw_name.find ("GStreamer") != -1:
			try:
				import pygst
				pygst.require ("0.10")
				import gst

				registry = gst.registry_get_default ()
				if registry.find_plugin ("pps") == None:
					"""This means an error will be displayed when they try to play anything"""
					self.gstreamer_plugins_present = False
				else:
					versions = registry.find_plugin ("pps").get_version ().split ('.')
					versions = [ int(v) for v in versions ]
					if len (versions) < 4:
						versions += [0] * (4 - len(versions))
					### 0.0.17.10 or later is ok
					if versions[0] > 0 or versions[1] > 0 or \
					   versions[2] > 17 or versions[3] >= 10:
						self.can_recode_mrl = True
			except ImportError:
				"""Do nothing; either it's using xine or python-gstreamer isn't installed"""

		"""Continue loading the plugin as before"""
		self.builder = self.load_ui ("ppstream.ui", True, totem_object.get_main_window (), self)
		self.config_builder = self.load_ui ("ppstream-config.ui", True,
						    totem_object.get_main_window (), self)

		self.totem = totem_object

		self.search_entry = self.builder.get_object ("pps_search_entry")
		self.search_entry.connect ("activate", self.on_search_entry_activated)
		self.search_button = self.builder.get_object ("pps_search_button")
		self.search_mode_button = self.builder.get_object ('pps_search_checkbutton')
		self.search_mode_button.set_label (_("filter:"))
		self.search_mode_button.connect ('toggled', self.on_search_mode_entry_toggled)

		#self.search_button.set_label ('')
		self.search_button.connect ("clicked", self.on_search_button_clicked)
		self.refresh_button = self.builder.get_object ("pps_refresh_button")
		#self.refresh_button.set_label ('')
		self.refresh_button.connect ("clicked", self.on_refresh_button_clicked)
		self.progress_bar = self.builder.get_object ("pps_progress_bar")
		self.progress_bar.hide ()

		self.notebook = self.builder.get_object ("pps_notebook")
		self.notebook.connect ("switch-page", self.on_notebook_page_changed)

		self.notebook_pages = ["categories", "movies", "files"]
		self.current_treeview_name = "categories"
		self.load_config ()
		self.setup_favorites ()
		self.setup_recent ()
		self.setup_search ()
		self.setup_categories ()
		self.setup_movies ()
		self.setup_files ()
		self.setup_page_ui ()
		self.setup_config_dialog ()

		self.vbox = self.builder.get_object ("pps_vbox")
		self.vbox.show_all ()
		self.page_hbox.hide()
		self.favorites_hbox.hide()

		totem_object.add_sidebar_page ("ppstream", _("PPStream"), self.vbox)

		self.retrieveimage = RetrieveImage ()
		self.retrieveimage.start ()

		self.search_terms = ''
		self.search_web_terms = ''
		self.search_web = False
		self.classes = []
		self.fetch_classes ()

		self.start_idle_task ()

	def deactivate (self, totem):
		self.clear_ui ()
		self.categories_count +=1
		self.movies_count += 1
		self.classes = []

		self.retrieveimage.cancel ()
		self.retrieveimage.join ()

		self.stop_idle_task ()

		self.ppslist.reset()

		totem.remove_sidebar_page ("ppstream")

	def is_configurable (self):
		return True

	def create_configure_dialog (self):
		dialog = self.config_builder.get_object ("pps_config_dialog")
		button = self.config_builder.get_object ("pps_config_show_posters")
		button.set_active (self.configs['show_posters'])
		button = self.config_builder.get_object ("pps_config_compatible")
		button.set_active (self.configs['compatible'])
		return dialog

	def start_idle_task (self):
		if self.configs['compatible']:
			self.source_id = gobject.idle_add (self.on_idle)
		else:
			self.source_id = 0

	def stop_idle_task (self):
		if self.source_id:
			gobject.source_remove (self.source_id)
			self.source_id = 0

	def on_idle (self):
		time.sleep (0.001)
		return True

	def load_special_classes (self):
		### show favorites
		self.subclass_index += 1
		self.ppslist.updateClasses([self.favorites])

		### show recent
		self.subclass_index += 1
		self.ppslist.updateClasses([self.recent])

		### search
		self.subclass_index += 1
		self.ppslist.updateClasses([self.search])

		### fill loaded categories
		self.fill_all_categories ()

	def __fetch_classes (self, *args):
		self.fetching_classes = True
		return self.ppslist.fetchClasses()

	def fetch_classes (self, force = False):
		if force:
			self.ppslist.reset()

		if self.ppslist.getClasses() == 0:
			return

		self.movie_class_path = ()
		self.subclass_index = 0
		self.categories_count += 1
		self.movies_count += 1
		self.fetching_classes = False
		self.clear_ui ()
		self.retrieveimage.cancel_tasks ()

		### load special classes first
		self.load_special_classes ()

		### download classes from pps.tv
		thread = ProcessThread (None,
					self.__fetch_classes,
					self.on_classes_fetched,
					self.categories_count)
		thread.start()

		gobject.timeout_add (300,
				     self.update_categories_progress,
				     self.categories_count)

	def on_classes_fetched (self, res, count):
		self.fetching_classes = False
		if not res:
			print "Couldn't load ppstream playlist"
			return

		### check if this is a canceled task
		if self.categories_count == count:
			self.parse_class (res)

	def __parse_movies (self, res, class_path, page, count):
		cls = self.get_movie_class (class_path)

		if res:
			try:
				result = self.ppslist.parseMovieList(cls, res, page)
			except Exception, e:
				print "Couldn't parse movie list:", class_path, e
				result = None

		cls.parsing_movies = False
		return result

	def on_movies_fetched (self, res, class_path, page, count, new = True):
		if self.movies_count != count:
			return

		### check if the server repeats the last page
		cls = self.get_movie_class (class_path)

		if res is None:
			cls.fetching_movies = False
			return

		### If the size of content is larger then 4k, parse it in another thread.
		### Blocking the main thread is bad.
		if res and len(res) > 1024 * 2:
			cls.parsing_movies = True
			process = ProcessThread (res, self.__parse_movies,
						 self.add_parsed_movies,
						 class_path, page, count)
			process.start ()
			self.update_progress_bar ()
			return

		result = self.__parse_movies (res, class_path, page, count)
		self.add_parsed_movies (result, class_path, page, count)

	def add_parsed_movies (self, result, class_path, page, count):
		if self.movies_count != count or page != self.movies_page:
			return
		cls = self.get_movie_class (class_path)

		if result:
			cls.addMovies(result, page)

		cls.fetching_movies = False

		if result:
			## for movie in result:
			## 	print movie
			self.fill_movies (result, class_path)
		self.set_page_max(cls.getMaxPage())

	def parse_class (self, res):
		### parse the class list
		classes = self.ppslist.parseClasses(res)
		for cls in classes:
			cls.fetching_movies = False
		self.ppslist.updateClasses(classes)

		### fill category list
		self.clear_pages ('categories')
		for cls in self.ppslist.getClasses():
			self.fill_categories (cls.id)

	def get_movie_class (self, class_path):
		def find_class (classes, clsid):
			for cls in classes:
				if cls.id == clsid:
					return cls
			return None
		if not class_path:
			return None
		classes = self.ppslist.getClasses()
		return find_class (classes, class_path[0])

	def download_movies (self, callback, class_path, page, count):
		cls = self.get_movie_class (class_path)
		thread = ProcessThread (None, self.__fetch_movies,
					callback, class_path,
					page, count)
		thread.start()

	def __fetch_movies(self, dummy, class_path, page, *args):
		cls = self.get_movie_class (class_path)
		if cls.id != CATEGORY_ID_SEARCH:
			return self.ppslist.fetchMovieList(cls, page)
		return self.ppslist.searchMovieList(cls.keyword, page)

	def fetch_movies (self, class_path):
		cls = self.get_movie_class (class_path)
		cls.fetching_movies = True
		#print cls, cls.page
		self.download_movies (self.on_movies_fetched, class_path,
				     self.movies_page, self.movies_count)

	def on_category_row_activated(self, treeview, path, view_column,
				      data = None):
		model, rows = treeview.get_selection ().get_selected_rows ()
		iter = model.get_iter (rows[0])
		title = model.get_value (iter, 3)
		classid = model.get_value (iter, 0)

		class_path = (classid,)
		#print title, class_path

		self.movies_page = 0
		self.movies_page = 0
		self.show_movies (class_path)
		self.notebook.set_current_page (1)

	def setup_categories (self):
		treeview_name = 'categories'
		treeview = self.builder.get_object ("pps_treeview_" + treeview_name)
		renderer = gtk.CellRendererText()
		renderer.set_property("xalign", 0.0)

		column = gtk.TreeViewColumn(_("Name"), renderer, text = 3)
		column.set_clickable(True)
		treeview.append_column(column)

		column = gtk.TreeViewColumn(_("Number"), renderer, text = 4)
		treeview.append_column(column)

		self.vadjust[treeview_name] = treeview.get_vadjustment ()
		self.vadjust[treeview_name].connect ("value-changed", self.on_value_changed)
		vscroll = self.builder.get_object ("pps_scrolled_window_" + treeview_name).get_vscrollbar ()
		vscroll.connect ("button-press-event", self.on_button_press_event)
		vscroll.connect ("button-release-event", self.on_button_release_event)

		self.liststore[treeview_name] = self.builder.get_object ("pps_liststore_" + treeview_name)
		self.treeview[treeview_name] = treeview
		treeview.set_model (self.liststore[treeview_name])

		treeview.connect("row_activated", self.on_category_row_activated)

		self.categories_count = 0

	def on_add_to_favorites (self, button):
		selection = self.treeview['movies'].get_selection ()
		model, rows = selection.get_selected_rows ()
		if not rows:
			return

		iter = model.get_iter (rows[0])
		class_path = eval(model.get_value (iter, 4))

		for row in rows:
			iter = model.get_iter (row)
			index = rows
			if type (model) is gtk.TreeModelFilter:
				iter = model.convert_iter_to_child_iter (iter)
				filtermodel = model.get_model ()
			else:
				filtermodel = model
			class_path = eval(filtermodel.get_value (iter, 4))
			self.add_favorites (class_path,
					    filtermodel.get_path (iter)[0], False)
		self.save_favorites ()
		self.refresh_favorites ()

	def on_remove_from_favorites (self, button):
		selection = self.treeview['movies'].get_selection ()
		model, rows = selection.get_selected_rows ()
		if not rows:
			return
		rows.reverse ()
		for row in rows:
			iter = model.get_iter (row)
			index = rows
			if type (model) is gtk.TreeModelFilter:
				iter = model.convert_iter_to_child_iter (iter)
				filtermodel = model.get_model ()
			else:
				filtermodel = model
			class_path = eval(filtermodel.get_value (iter, 4))
			self.remove_favorites (class_path,
					       filtermodel.get_path (iter)[0], False)
		self.save_favorites ()
		self.refresh_favorites ()

	def setup_favorites (self):
		button = self.builder.get_object ('pps_add_to_favorites_button')
		self.add_to_favorites_button = button
		button.set_sensitive (False)
		button.connect ('clicked', self.on_add_to_favorites)
		button = self.builder.get_object ('pps_remove_from_favorites_button')
		button.set_sensitive (False)
		button.connect ('clicked', self.on_remove_from_favorites)
		self.remove_from_favorites_button = button
		self.favorites_hbox = self.builder.get_object ('pps_favorites_hbox')
		self.load_favorites ()

	def load_favorites (self):
		self.favorites = ppslist.PPSClass ()
		self.favorites.id = CATEGORY_ID_FAVORITES
		self.favorites.title = _('Favorites')
		self.favorites.fetching_movies = False

		filename = os.path.join (pps_cache_dir, 'favorites.pickle')
		if not os.path.exists (filename):
			return
		try:
			dump = pickle.load (file(filename, 'rb'))
		except Exception, e:
			print "Couldn't load favorite list", e
			dump = []

		movies = []
		for d in dump:
			movie = ppslist.PPSMovie()
			movie.load(d)
			movies.append(movie)
		self.favorites.setMovies(movies)

	def save_favorites (self):
		treeview_name = 'categories'
		filename = os.path.join (pps_cache_dir, 'favorites.pickle')
		fp = None
		try:
			fp = file (filename, 'wb')
		except:
			return
		d = [ movie.dictionary() for movie in self.favorites.getPage(0)]
		try:
			pickle.dump(d, fp)
		except:
			pass

	def add_favorites (self, class_path, index, update = True):
		cls = self.get_movie_class (class_path)
		movies = cls.getPage(0)
		movie = movies[index]
		if movie in self.favorites.getPage(0):
			return
		movie = copy.copy(movie)
		movie.model_path = ()
		self.favorites.addMovies([movie])

		if not update:
			return
		self.refresh_favorites ()
		self.save_favorites ()

	def remove_favorites (self, class_path, index, update = False):
		cls = self.get_movie_class (class_path)
		movies = cls.getPage(self.movies_page)
		movie = movies[index]
		#print self.favorites.movies, movie
		if not movie in self.favorites.getPage(0):
			print "Couldn't find ", movie, ' in favorites'
			return
		movies.remove(movie)

		if not update:
			return
		self.refresh_favorites ()
		self.save_favorites ()

	def refresh_favorites (self):
		treeview_name = 'categories'
		treeview = self.builder.get_object ("pps_liststore_" + treeview_name)
		if not len (treeview):
			return

		self.update_category (self.favorites)

		if self.movie_class_path == ():
			return

		class_path = self.movie_class_path
		if class_path != (CATEGORY_ID_FAVORITES,):
			return

		self.movie_class_path = ()
		self.show_movies (class_path, False, True)

	def setup_recent (self):
		self.load_recent ()

	def load_recent (self):
		treeview_name = 'categories'

		self.recent = ppslist.PPSClass ()
		self.recent.id = CATEGORY_ID_RECENT;
		self.recent.title = _('Recent')
		self.recent.fetching_movies = False

		filename = os.path.join (pps_cache_dir, 'recent.pickle')
		if not os.path.exists (filename):
			return
		try:
			dump = pickle.load (file(filename, 'rb'))
		except Exception, e:
			print "Couldn't load recent list", e
			dump = []

		movies = []
		for d in dump:
			movie = ppslist.PPSMovie()
			movie.load(d)
			movies.append(movie)
		self.recent.setMovies(movies)

	def save_recent (self):
		treeview_name = 'categories'
		filename = os.path.join (pps_cache_dir, 'recent.pickle')
		fp = None
		try:
			fp = file (filename, 'wb')
		except:
			return
		d = [ movie.dictionary() for movie in self.recent.getPage(0)]
		try:
			pickle.dump(d, fp)
		except:
			pass

	def add_recent (self, class_path, index):
		cls = self.get_movie_class (class_path)
		movie = cls.getPage(self.movies_page)[index]

		liststore_name = 'movies'
		liststore = self.liststore[liststore_name]
		old_index = -1
		movies = self.recent.getPage(0)
		if movie in movies:
			### it's alreay the last one
			if movie == movies[0]:
				return
			old_index = movies.index (movie)
			movies.remove (movie)
		else:
			while len (movies) > MAX_RECENT - 1:
				if self.movie_class_path == (CATEGORY_ID_RECENT,):
					del self.liststore[-1]
				del movies[-1]
		self.recent.resetMovies()
		movie = copy.copy(movie)
		self.recent.setMovies([movie] + movies)

		if self.movie_class_path == (CATEGORY_ID_RECENT,):
			if old_index < 0:
				self.storelist_append_movie (liststore, movie,
							     self.movie_class_path, 0)
			else:
				iter = liststore.get_iter (old_index)
				liststore.move_after (iter, None)
		self.refresh_recent_num ()
		self.save_recent ()

	def refresh_recent_num (self):
		liststore_name = 'categories'
		liststore = self.liststore[liststore_name]
		if not len (liststore):
			return

		self.update_category (self.recent)

	def refresh_recent (self):
		self.refresh_recent_num ()
		if self.movie_class_path == ():
			return

		class_path = self.movie_class_path
		if class_path != (CATEGORY_ID_RECENT,):
			return

		self.movie_class_path = ()
		self.show_movies (class_path, False, True)

	def setup_search (self):
		self.search = ppslist.PPSClass()
		self.search.id = CATEGORY_ID_SEARCH
		self.search.title = _('Search result')
		self.search.keyword = ''
		self.search.fetching_movies = False

	def fill_all_categories (self):
		treeview_name = 'categories'
		treeview = self.liststore[treeview_name]
		self.clear_pages (treeview_name)
		for cls in self.ppslist.getClasses():
			iter = treeview.append(None)
			s = '%d/%d' % (len(cls.getMovies()), cls.getMaxPage())
			treeview.set(iter,
				     0, cls.id,
				     1, 0,
				     2, 0,
				     3, cls.title,
				     4, s)
			cls.model_path = treeview.get_path(iter)

	def fill_categories (self, clsid):
		treeview_name = 'categories'
		treeview = self.liststore[treeview_name]

		cls = None
		for cls in self.ppslist.getClasses():
			if cls.id == clsid:
				break
		if not cls:
			return

		s = '%d/%d' % (len(cls.getMovies()), cls.getMaxPage())
		iter = treeview.append(None)
		treeview.set(iter,
			     0, cls.id,
			     1, 0,
			     2, 0,
			     3, cls.title,
			     4, s)
		cls.model_path = treeview.get_path(iter)
		#print cls.id, cls.title

	def update_category (self, cls):
		treeview_name = 'categories'
		liststore = self.liststore[treeview_name]
		iter = liststore.get_iter (cls.model_path)
		s = '%d/%d' % (len(cls.getMovies()), cls.getMaxPage())
		liststore.set (iter, 4, s)

	def update_current_category (self):
		if not self.movie_class_path:
			return
		cls = self.get_movie_class(self.movie_class_path)
		self.update_category (cls)

	def get_movie_tooltip (self, movie):
		tip = movie.title
		if movie.director:
			tip += '\n' + _("Director: ") + movie.director
		if movie.actor:
			tip += '\n' + _("Actor: ") + movie.actor
		if movie.area:
			tip += '\n' + _("Area: ") + movie.area
		if movie.size:
			tip += '\n' + _("Size: ") + str(movie.size) + _('MB')
		if movie.pubtime:
			tip += '\n' + _("Pubtime: ") + movie.pubtime
		if movie.length:
			tip += '\n' + _("Length: ") + str(movie.length) + _('Min')
		tip += '\n' + _("Score: ") + str(movie.score)
		tip += '\n' + _("Episodes: ") + str(movie.cn)
		tip += '\n' + _("Description: ") + '\n' + movie.desc
		if hasattr(file, 'longdesc'):
			tip += '\n' + _("Long Description: ") + movie.longdesc

		return unescape_xml (tip)

	def on_image_retrieved (self, pixbuf, movie, iter, class_path, count):
		treeview_name = 'movies'
		liststore = self.liststore[treeview_name]
		movie.pixbuf = pixbuf

		try:
			### check whether the treeview has been cleared.
			if self.movie_class_path == class_path and self.movies_count == count:
				if self.configs['show_posters']:
					liststore.set (iter, 0, pixbuf)
				else:
					print 'not showing posters'
		except Exception, e:
			print 'Failed to loading posters:', class_path, e
			pass

	def recode_mrl (self, url):
		if not self.can_recode_mrl:
			return url
		path, qs = urllib.splitquery (url)
		if '%' in path:
			path = urllib.unquote (path)
			try:
				path = unicode (path, 'gbk').encode('utf-8')
			except Exception, e:
				print 'Failed to recode url:', e
				return url
		if qs is None:
			qs = ''
		return path + qs

	def storelist_append_movie (self, treeview, movie, class_path, pos = -1):
		files = movie.getFiles(self.files_page)
		#if not movie.getFiles():
		#	return
		f = None
		if len(files):
			f = files[0]
			origurl = f.url
		else:
			origurl = ''
		tip = self.get_movie_tooltip (movie)
		pixbuf = movie.pixbuf
		if not self.configs['show_posters']:
			pixbuf = None
		url = self.recode_mrl (origurl)
		if pos == -1:
			it = treeview.append ([pixbuf, movie.title, url, tip,
					       repr(class_path), origurl])
		else:
			it = treeview.insert (0)
			treeview.set (it,
				      0, pixbuf,
				      1, movie.title,
				      2, url,
				      3, tip,
				      4, repr(class_path),
				      5, f.url)
		if not movie.pixbuf and movie.smallimage and self.configs['show_posters']:
			self.retrieveimage.retrieve (movie.smallimage, self.on_image_retrieved,
						     movie, it, class_path,
						     self.movies_count)
		movie.model_path = treeview.get_path(it)

	def show_files_current_page(self):
		movie = self.current_movie
		if self.current_treeview_name != 'files':
			return

		if movie:
			self.set_page_max(movie.getMaxPage())
			self.set_current_page(self.files_page + 1)
		else:
			self.set_page_max(1)
			self.set_current_page(1)

	def fill_files (self, movie):
		treeview_name = 'files'
		self.clear_pages (treeview_name)
		treeview = self.liststore[treeview_name]
		files = movie.getFiles(self.files_page)
		num_files = len(files)
		nr = 0
		for f in files:
			no = str(nr + 1) + '/' + str(num_files)
			it = treeview.append ([no, f.title, self.recode_mrl (f.url), f.url])
			nr += 1
		self.show_files_current_page()

	def show_files (self, movie):
		if self.current_movie and self.current_movie != movie:
			self.files_page = 0
		self.current_movie = movie
		if movie.getFiles(self.files_page):
			self.fill_files(movie)
		else:
			self.clear_pages('files')
			self.fetch_files(movie, self.files_page)
			self.show_files_current_page()

	def download_files (self, callback, movie, page, show, count):
		thread = ProcessThread (None, self.__fetch_files,
					callback, movie, page, show, count)
		thread.start()

	def __fetch_files(self, dummy, movie, page, show, count):
		cls = movie.parent
		assert (cls is not None)
		if not movie.baseurl:
			return self.ppslist.fetchMovie(cls, movie)
		return self.ppslist.fetchMoviePlayList(cls, movie, page)

	def fetch_files (self, movie, page = 0, show = True):
		self.download_files (self.on_files_fetched, movie,
				     page, show, self.movies_count)

	def on_files_fetched(self, res, movie, page, show, movies_count):
		curpage = page
		if res:
			if movie.baseurl:
				result = self.ppslist.parseMoviePlayList(movie.parent,
									movie, res,
									page)
				if not result:
					files = []
				else:
					files = result[0]
					curpage = result[2]
			else:
				files = self.ppslist.parseMovie(movie.parent, movie, res)
			if not files and not movie.baseurl:
				url = self.ppslist.parseMoviePlayListUrl(movie.parent,
									 movie, res)
				assert(curpage == 0)
                                ### the default page
				curpage = -1
				self.download_files(self.on_files_fetched, movie,
						    curpage, show, movies_count)
			else:
				if curpage != page and self.movies_count == movies_count and \
				   movie == self.current_movie:
					#print 'updating files_page to', curpage
					self.files_page = curpage

				movie.setFiles(files, curpage)
		if show:
			if self.movies_count == movies_count and \
			   movie == self.current_movie and curpage == self.files_page:
				self.fill_files(movie)
		self.movies_update_files_status()

	def fill_all_movies (self, class_path):
		cls = self.get_movie_class (class_path)

		self.set_page_max(cls.getMaxPage())

		treeview_name = 'movies'
		liststore = self.liststore[treeview_name]
		for movie in cls.getPage(self.movies_page):
			self.storelist_append_movie (liststore, movie, class_path)

		self.update_category (cls)

	def fill_movies (self, movies, class_path):
		if self.movie_class_path != class_path:
			return
		treeview_name = 'movies'
		liststore = self.liststore[treeview_name]
		for movie in movies:
			self.storelist_append_movie (liststore, movie, class_path)

		self.update_current_category ()

	def refetch_movies (self, class_path = None):
		if class_path is None:
			class_path = self.movie_class_path

		self.retrieveimage.cancel_tasks ()
		self.movie_class_path = ()
		self.show_movies (class_path, True)

	def show_movies (self, class_path, force = False, cancel = True):
		#print class_path, force, cancel
		### force -- refetch movie list from server
		### cancel -- discard the current movie list
		if self.movie_class_path == class_path and not force:
			return

		### cancel current fetching thread
		if cancel:
			self.movies_count += 1
			if self.movie_class_path != ():
				cls = self.get_movie_class (self.movie_class_path)
				cls.fetching_movies = False

		self.movie_class_path = class_path

		if not self.configs['show_posters']:
			self.retrieveimage.cancel_tasks ()

		self.clear_pages (['movies', 'files'])

		cls = self.get_movie_class (class_path)

		### reset movie list(except the favorites/recent list)
		if force:
			cls.setMovies([], self.movies_page)

		if cancel or (not cls.fetching_movies and \
			      not self.retrieveimage.is_busy ()):
			self.add_progress_callback (class_path,
						    self.movies_count)

		### fill all loaded movies
		self.set_current_page(self.movies_page + 1)
		self.fill_all_movies (class_path)
		if not len(cls.getPage(self.movies_page)) and \
		   not cls.fetching_movies:
			self.fetch_movies (class_path)

	def on_movies_visible_func (self, model, iter):
		terms = self.search_terms
		if not terms:
			return True
		treeview = self.treeview['movies']
		title = model.get (iter, treeview.get_property ("title-column"))[0] or ''
		tooltip = model.get (iter, treeview.get_property ("tooltip-column"))[0] or ''
		if terms in title or terms in tooltip:
			return True
		return False

	def setup_movies (self):
		treeview_name = 'movies'
		self.movie_class_path = ()
		self.movies_count = 0
		self.movies_page = 0

		"""This is done here rather than in the UI file,
		because UI files parsed in C and GObjects created in
		Python apparently don't mix."""
		renderer = totem.CellRendererVideo (use_placeholder = self.configs['show_posters'])
		treeview = PPSVideoList.PPSVideoList()
		treeview.set_property('headers-visible', False)
		treeview.set_property('fixed-height-mode', False)
                treeview.set_property('title-column', 1)
                treeview.set_property('tooltip-column', 3)
                treeview.set_property('mrl-column', 2)
		treeview.set_property ("totem", self.totem)

		window = self.builder.get_object('pps_scrolled_window_movies')
		window.add(treeview)

		treeview.connect ("row-activated", self.on_movies_row_activated)
		treeview.connect_after ("starting-video", self.on_movies_starting_video)
		treeview.insert_column_with_attributes (0, _("Videos"),
							renderer, thumbnail=0, title=1)
		self.cell_video_renderer = renderer

		self.vadjust[treeview_name] = treeview.get_vadjustment ()
		self.vadjust[treeview_name].connect ("value-changed", self.on_value_changed)
		vscroll = self.builder.get_object ("pps_scrolled_window_" + treeview_name).get_vscrollbar ()
		vscroll.connect ("button-press-event", self.on_button_press_event)
		vscroll.connect ("button-release-event", self.on_button_release_event)

		self.liststore[treeview_name] = self.builder.get_object ("pps_liststore_" + treeview_name)
		self.treeview[treeview_name] = treeview
		treefilter = self.liststore[treeview_name].filter_new ()
		treefilter.set_visible_func (self.on_movies_visible_func)
		treeview.set_model (treefilter)

		selection = treeview.get_selection ()
		selection.connect ("changed", self.on_movies_selection_changed)

	def on_add_to_playlist (self, treeview):
		### add selected movie to recent list
		treeview_name = 'movies'
		treeview = self.treeview[treeview_name]
		selection = treeview.get_selection ()
		model = treeview.get_model ()
		model, rows = selection.get_selected_rows ()
		self.add_movie_to_recent (treeview, rows[0])

		treeview_name = 'files'
		treeview = self.treeview[treeview_name]
		selection = treeview.get_selection ()
		model = treeview.get_model ()
		model, rows = selection.get_selected_rows ()

		for row in rows:
			iter = model.get_iter (row)
			mrl = model.get_value (iter, 2)
			self.totem.action_remote (totem.REMOTE_COMMAND_ENQUEUE, mrl)

	def on_copy_location (self, action):
		treeview_name = 'files'
		treeview = self.treeview[treeview_name]
		selection = treeview.get_selection ()
		model = treeview.get_model ()
		model, rows = selection.get_selected_rows ()
		iter = model.get_iter (rows[0])
		mrl = model.get_value (iter, 3)

		clip = gtk.Clipboard ()
		clip.set_text (mrl)
		clip = gtk.Clipboard (display = gtk.gdk.display_get_default(),
				      selection = "PRIMARY")
		clip.set_text (mrl)

	def setup_files_action (self):
		datadir = os.path.dirname (__file__)
                filename = os.path.join (datadir, 'ppstream-ui.xml')

		self.files_uimanager = self.builder.get_object ('pps-list-ui-manager')
		self.files_uimanager.add_ui_from_file(filename)

		action_group = gtk.ActionGroup(name = 'pps-list-action-group')
		self.files_uimanager.insert_action_group(action_group, 0)

		action = gtk.Action(name = 'add-to-playlist',
				    label = _('_Add to Playlist'),
				    tooltip = _('Add the video to the playlist'),
				    stock_id = 'gtk-add')
		action.connect ('activate', self.on_add_to_playlist)
		action_group.add_action(action)

		action = gtk.Action(name = 'copy-location',
				    label = _('_Copy Location'),
				    tooltip = _('Copy the location to the clipboard'),
				    stock_id = 'gtk-copy')
		action.connect ('activate', self.on_copy_location)
		action_group.add_action(action)

	def files_show_popup_menu (self, event = None):
		treeview_name = 'files'
		treeview = self.treeview [treeview_name]
		selection = treeview.get_selection ()

		if event:
			button = event.button
			time = event.time
			path = treeview.get_path_at_pos (int(event.x), int(event.y))
			if not path or not selection.path_is_selected (path[0]):
				selection.unselect_all ()
		else:
			time = gtk.get_current_event_time ()
			button = None

		count = selection.count_selected_rows ()
		if not count:
			return False

		action_group = self.files_uimanager.get_action_groups () [0]
		action = action_group.get_action ("copy-location")
		action.set_sensitive (count == 1)

		menu = self.files_uimanager.get_widget ("/pps-list-popup")
		menu.select_first (False)
		menu.popup (None, None, None, button, time)

		return True

	def on_files_popup_menu (self):
		self.files_show_popup_menu ()

	def on_files_button_pressed (self, widget, event, user_data = None):
		if event.type == gtk.gdk.BUTTON_PRESS and event.button == 3:
			return self.files_show_popup_menu (event)

		return False

	def setup_files (self):
		treeview_name = 'files'
		treeview = self.builder.get_object ("pps_treeview_" + treeview_name)
		renderer = gtk.CellRendererText()
		renderer.set_property("xalign", 0.0)

		column = gtk.TreeViewColumn(_("No"), renderer, text = 0)
		column.set_clickable(True)
		treeview.append_column(column)

		column = gtk.TreeViewColumn(_("Title"), renderer, text = 1)
		treeview.append_column(column)

		column = gtk.TreeViewColumn(_("Mrl"), renderer, text = 2)
		treeview.append_column(column)

		self.vadjust[treeview_name] = treeview.get_vadjustment ()
		self.vadjust[treeview_name].connect ("value-changed", self.on_value_changed)
		vscroll = self.builder.get_object ("pps_scrolled_window_" + treeview_name).get_vscrollbar ()
		vscroll.connect ("button-press-event", self.on_button_press_event)
		vscroll.connect ("button-release-event", self.on_button_release_event)

		self.liststore[treeview_name] = self.builder.get_object ("pps_liststore_" + treeview_name)
		self.treeview[treeview_name] = treeview
		treeview.set_model (self.liststore[treeview_name])

		treeview.connect("row_activated", self.on_files_treeview_row_activated)

		self.setup_files_action ()
		treeview.connect ("popup-menu", self.on_files_popup_menu)
		treeview.connect ("button-press-event", self.on_files_button_pressed)

		selection = treeview.get_selection ()
		selection.set_mode (gtk.SELECTION_MULTIPLE)

		self.files_page = 0
		self.files_count = 0

	def on_files_treeview_row_activated(self, treeview, path, view_column, data=None):
		model, rows = treeview.get_selection ().get_selected_rows ()
		try:
			iter = model.get_iter (rows[0])
		except IndexError:
			print 'activated invalid index:', rows[0]
			return

		### play the movie
		title = model.get_value (iter, 1)
		mrl = model.get_value (iter, 2)
		#print title, mrl
		if hasattr(self.totem, "add_to_playlist_and_play"):
			self.totem.add_to_playlist_and_play (mrl, title, False)
		else:
			self.totem.action_remote (totem.REMOTE_COMMAND_REPLACE, mrl)

		### add selected movie to recent list
		movies_treeview = self.treeview['movies']
		movies_model, movies_rows = movies_treeview.get_selection ().get_selected_rows ()
		self.add_movie_to_recent (movies_treeview, movies_rows[0])

	def on_movies_selection_changed (self, selection):
		model, rows = selection.get_selected_rows ()
		if not rows:
			self.clear_pages ('files')
			self.current_movie = None
			self.files_page = 0
			self.show_files_current_page()
			return

		all_files_fetched = True
		for row in rows:
			iter = model.get_iter(row)
			if type (model) is gtk.TreeModelFilter:
				iter = model.convert_iter_to_child_iter (iter)
				model = model.get_model ()

			class_path = eval(model.get_value (iter, 4))
			cls = self.get_movie_class (class_path)

			index = model.get_path(iter)[0]
			movie = cls.getPage(self.movies_page)[index]
			if not movie.getFiles(self.files_page):
				all_files_fetched = False
				if row != rows[0]:
					self.fetch_files(movie, False)
			if row == rows[0]:
				self.show_files(movie)
		if self.movie_class_path == (CATEGORY_ID_FAVORITES,):
			self.add_to_favorites_button.set_sensitive(False)
		else:
			self.add_to_favorites_button.set_sensitive(all_files_fetched)

	def movies_update_files_status(self):
		selection = self.treeview['movies'].get_selection ()
		model, rows = selection.get_selected_rows ()
		if not rows:
			self.clear_pages ('files')
			return

		all_files_fetched = True
		for row in rows:
			iter = model.get_iter(row)
			if type (model) is gtk.TreeModelFilter:
				iter = model.convert_iter_to_child_iter (iter)
				model = model.get_model ()

			class_path = eval(model.get_value (iter, 4))
			cls = self.get_movie_class (class_path)

			index = model.get_path(iter)[0]
			movie = cls.getPage(self.movies_page)[index]
			if not movie.getFiles(self.files_page):
				all_files_fetched = False
			else:
				url = model.get(iter, 2)[0]
				if not url:
					url = movie.getFiles(self.files_page)[0].url
					model.set(iter, 2, self.recode_mrl(url))
					model.set(iter, 5, url)

		if self.movie_class_path == (CATEGORY_FAVORITES,):
			self.add_to_favorites_button.set_sensitive(False)
		else:
			self.add_to_favorites_button.set_sensitive(all_files_fetched)

	def movies_goto_first_page(self):
		cls = self.get_movie_class(self.movie_class_path)

		if self.movies_page == 0:
			return
		self.movies_page = 0

		self.show_movies(self.movie_class_path, force = True)

	def files_goto_first_page(self):
		if self.files_page == 0:
			return
		self.files_page = 0
		self.show_files(self.current_movie)

	def on_first_page_clicked(self, *args):
		if self.movie_class_path == ():
			return
		if self.current_treeview_name == 'movies':
			self.movies_goto_first_page()
		elif self.current_treeview_name == 'files':
			self.files_goto_first_page()

	def movies_goto_prev_page(self):
		if self.movies_page < 1:
			return
		self.movies_page -= 1
		self.show_movies(self.movie_class_path, force = True)

	def files_goto_prev_page(self):
		if self.files_page < 1:
			return
		self.files_page -= 1
		self.show_files(self.current_movie)

	def on_prev_page_clicked(self, *args):
		if self.movie_class_path == ():
			return
		if self.current_treeview_name == 'movies':
			self.movies_goto_prev_page()
		elif self.current_treeview_name == 'files':
			self.files_goto_prev_page()

	def movies_goto_next_page(self):
		cls = self.get_movie_class(self.movie_class_path)

		if self.movies_page >= cls.getMaxPage() - 1:
			return
		self.movies_page += 1

		self.show_movies(self.movie_class_path, force = True)

	def files_goto_next_page(self):
		if self.files_page >= self.current_movie.getMaxPage() - 1:
			return
		self.files_page += 1
		self.show_files(self.current_movie)

	def on_next_page_clicked(self, *args):
		if self.movie_class_path == ():
			return
		if self.current_treeview_name == 'movies':
			self.movies_goto_next_page()
		elif self.current_treeview_name == 'files':
			self.files_goto_next_page()

	def movies_goto_last_page(self):
		cls = self.get_movie_class(self.movie_class_path)

		if self.movies_page == cls.getMaxPage() - 1:
			return
		self.movies_page = cls.getMaxPage() - 1

		self.show_movies(self.movie_class_path, force = True)

	def files_goto_last_page(self):
		if self.files_page == self.current_movie.getMaxPage() - 1:
			return
		self.files_page = self.current_movie.getMaxPage() - 1
		self.show_files(self.current_movie)

	def on_last_page_clicked(self, *args):
		if self.movie_class_path == ():
			return
		if self.current_treeview_name == 'movies':
			self.movies_goto_last_page()
		elif self.current_treeview_name == 'files':
			self.files_goto_last_page()

	def movies_goto_page(self, page):
		cls = self.get_movie_class(self.movie_class_path)
		if page >= cls.getMaxPage():
			return
		if page == self.movies_page:
			return
		self.movies_page = page
		self.show_movies(self.movie_class_path, force = True)

	def files_goto_page(self, page):
		movie = self.current_movie
		if page >= movie.getMaxPage():
			return
		if page == self.files_page:
			return
		self.files_page = page
		self.show_files(movie)

	def on_goto_page_clicked(self, *args):
		if self.movie_class_path == ():
			return

		page = self.which_page_button.get_value_as_int() - 1
		if self.current_treeview_name == 'movies':
			self.movies_goto_first_page()
		elif self.current_treeview_name == 'files':
			self.files_goto_first_page()

	def set_page_max(self, max_page):
		adjustment = self.which_page_button.get_adjustment()
		adjustment.upper = max_page
		self.max_page_label.set_text (_('Total: %d') % max_page)

	def set_current_page(self, current):
		adjustment = self.which_page_button.get_adjustment()
		adjustment.value = current

	def setup_page_ui (self):
		self.page_hbox = self.builder.get_object('pps_page_hbox')
		self.first_page_button = self.builder.get_object('pps_first_page_button')
		self.prev_page_button = self.builder.get_object('pps_prev_page_button')
		self.next_page_button = self.builder.get_object('pps_next_page_button')
		self.last_page_button = self.builder.get_object('pps_last_page_button')
		self.goto_page_button = self.builder.get_object('pps_goto_page_button')
		self.which_page_button = self.builder.get_object('pps_which_page_button')
		self.max_page_label = self.builder.get_object('pps_max_page_label')

		self.page_hbox.set_sensitive(False)
		self.set_page_max (1)

		adjustment = self.which_page_button.get_adjustment()
		adjustment.lower = 1
		adjustment.upper = 1
		adjustment.value = 1

		self.first_page_button.connect('clicked', self.on_first_page_clicked)
		self.prev_page_button.connect('clicked', self.on_prev_page_clicked)
		self.next_page_button.connect('clicked', self.on_next_page_clicked)
		self.last_page_button.connect('clicked', self.on_last_page_clicked)
		self.goto_page_button.connect('clicked', self.on_goto_page_clicked)

	def on_notebook_page_changed (self, notebook, notebook_page, page_num):
		self.current_treeview_name = self.notebook_pages[page_num]
		self.add_to_favorites_button.set_sensitive (False)
		self.remove_from_favorites_button.set_sensitive (False)
		if self.current_treeview_name == 'files':
			self.refresh_button.set_sensitive (False)
			self.page_hbox.show()
			if self.current_movie:
				self.page_hbox.set_sensitive (True)
			else:
				self.page_hbox.set_sensitive (False)
			self.show_files_current_page()
			self.favorites_hbox.hide()
			self.update_progress_bar ()
		elif self.current_treeview_name == 'movies':
			if self.movie_class_path == ():
				self.refresh_button.set_sensitive (False)
			elif self.movie_class_path == (CATEGORY_ID_FAVORITES,):
				self.remove_from_favorites_button.set_sensitive (True)
			else:
				self.refresh_button.set_sensitive (True)
				self.page_hbox.set_sensitive (True)
			self.page_hbox.show ()
			self.favorites_hbox.show ()
			self.movies_update_files_status()
			self.update_progress_bar ()
			if self.movie_class_path:
				cls = self.get_movie_class(self.movie_class_path)
				self.set_current_page(self.movies_page + 1)
				self.set_page_max(cls.getMaxPage())
			else:
				self.set_current_page(1)
				self.set_page_max(1)
		else:
			self.page_hbox.set_sensitive (False)
			self.page_hbox.hide ()
			self.favorites_hbox.hide ()
			self.refresh_button.set_sensitive (True)
			self.update_categories_progress (self.categories_count)

	def update_categories_progress (self, count):
		if count != self.categories_count:
			return False
		if self.current_treeview_name != 'categories':
			return True
		classes = self.ppslist.getClasses()
		if self.fetching_classes:
			self.progress_bar.pulse ()
			self.progress_bar.set_text (_("Refreshing category list..."))
			self.progress_bar.show ()
			return True
		else:
			self.progress_bar.set_fraction (0.0)
			self.progress_bar.set_text ("")
			self.progress_bar.hide ()
			return False
		return True

	def update_progress_bar (self):
		if self.current_treeview_name == 'categories':
			return True
		if self.movie_class_path == ():
			self.progress_bar.set_fraction (0.0)
			self.progress_bar.set_text ("")
			self.progress_bar.hide ()
			return True

		cls = self.get_movie_class (self.movie_class_path)
		if cls.fetching_movies:
			self.progress_bar.pulse()
			progressstr = '(%d/%d)' % (self.movies_page + 1, cls.max_page)
			if hasattr(cls, 'parsing_movies') and cls.parsing_movies:
				self.progress_bar.set_text (_("Parsing movie list%s...") % progressstr)
			else:
				self.progress_bar.set_text (_("Loading movie list%s...") % progressstr)
			self.progress_bar.show ()
			return True
		elif self.retrieveimage.is_busy ():
			remaining = self.retrieveimage.get_remaning_tasks ()
			num_movies = len(cls.getMovies())
			if remaining > num_movies:
				remaining = num_movies
			finished = num_movies - remaining
			progress = float(finished) / num_movies
			self.progress_bar.set_fraction (progress)
			progressstr = str(finished) + '/' + str(num_movies)
			self.progress_bar.set_text (_('loading posters(%s)...') % progressstr)
			self.progress_bar.show ()
		else:
			self.progress_bar.set_fraction (0.0)
			self.progress_bar.set_text ("")
			self.progress_bar.hide ()

			return False

		return True

	def on_update_progress_bar (self, class_path, count):
		if self.movie_class_path == class_path and self.movies_count == count:
			return self.update_progress_bar ()
		return False

	def add_progress_callback (self, class_path, count):
		gobject.timeout_add (350, self.on_update_progress_bar,
				     class_path, count)

	def on_movies_row_activated (self, treeview, path, column):
		pass

	def add_movie_to_recent (self, treeview, path):
		model = treeview.get_model ()
		iter = model.get_iter (path)

		if type (model) is gtk.TreeModelFilter:
			iter = model.convert_iter_to_child_iter (iter)
			model = model.get_model ()

		class_path = eval(model.get_value (iter, 4))
		cls = self.get_movie_class(class_path)

		movie = cls.getPage(self.movies_page)[path[0]]
		files = movie.getFiles(self.files_page)
		if not files:
			return False

		self.add_recent (class_path, model.get_path (iter)[0])
		return True

	def on_movies_starting_video (self, widget, treeview, path):
		"""Display an error if the required GStreamer plugins aren't installed"""
		if self.gstreamer_plugins_present == False:
			self.totem.interface_error_with_link
			(_("Totem cannot play this type of media (%s) because you do not have the appropriate plugins to handle it.") % _("Ppstream"),
			 _("Please install the necessary plugins and restart Totem to be able to play this media."),
			 "http://www.gnome.org/projects/totem/#codecs",
			 _("More information about media plugins"),
			 self.totem.get_main_window ())
			return False

		return self.add_movie_to_recent (treeview, path)

	def on_button_press_event (self, widget, event):
		self.button_down = True

	def on_button_release_event (self, widget, event):
		self.button_down = False
		self.on_value_changed (self.vadjust[self.current_treeview_name])

	def on_value_changed (self, adjustment):
		"""Load more results when we get near the bottom of the treeview"""
		pass

	def clear_ui (self, what = None):
		#window = self.vbox.window
		#window.set_cursor (None)
		self.progress_bar.set_fraction (0.0)
		self.progress_bar.set_text ("")
		self.clear_pages (what)

	def clear_pages (self, what = None):
		if what is None:
			what = self.notebook_pages
		if not type(what) is list:
			what = [what]

		for treeview_name in what:
			treeview = self.treeview[treeview_name]
			treeview.get_selection ().unselect_all ()
			liststore = self.liststore[treeview_name]
			liststore.clear ()

	def refresh_categories (self):
		self.fetch_classes (True)

	def refresh_movies (self):
		if self.movie_class_path != ():
			self.refetch_movies ()

	def set_show_posters (self, show = True):
		self.cell_video_renderer.set_property ('use_placeholder', show)
		if self.movie_class_path == ():
			return
		class_path = self.movie_class_path
		self.movie_class_path = ()
		self.show_movies (class_path, False, False)

	def sync_config (self):
		for key in self.input_configs.keys ():
			keyname = gconf_key + '/' + key
			value = self.input_configs [key]
			if type (value) is bool:
				self.gconf.set_bool (keyname, value)
			elif type (value) is int:
				self.gconf.set_int (keyname, value)
			elif type (value) is float:
				self.gconf.set_float (keyname, value)
			else:
				self.gconf.set_string (keyname, value)
		if self.input_configs['show_posters'] != self.configs['show_posters']:
			self.configs['show_posters'] = self.input_configs['show_posters']
			self.set_show_posters (self.configs['show_posters'])

		if self.input_configs['compatible'] != self.configs['compatible']:
			self.configs['comptiable'] = self.input_configs['compatible']
			self.stop_idle_task ()
			self.start_idle_task ()

	def on_refresh_button_clicked (self, button):
		if self.current_treeview_name == 'categories':
			self.refresh_categories ()
		elif self.current_treeview_name == 'movies':
			self.refresh_movies ()

	def on_search_mode_entry_toggled (self, button):
		self.search_web = self.search_mode_button.get_active()
		if self.search_web:
			button.set_label (_('search:'))
		else:
			button.set_label (_("filter:"))

	def on_search_entry_activated (self, entry):
		if self.search_web:
			self.search.keyword = self.search_entry.get_text ()
			self.search_terms = ''
			self.movies_page = 0
			self.files_page = 0
			self.show_movies ((CATEGORY_ID_SEARCH,), True, True)
			self.notebook.set_current_page (1)
		else:
			self.search_terms = self.search_entry.get_text ()
			self.treeview['movies'].get_model ().refilter ()

	def on_search_button_clicked (self, button):
		self.on_search_entry_activated (self.search_entry)

	def on_config_dialog_delete (self, dialog, event):
		dialog.hide ()
		return True

	def on_config_dialog_close (self, dialog):
		dialog.hide ()

	def on_config_dialog_response (self, dialog, response_id):
		dialog.hide ()
		if response_id == gtk.RESPONSE_DELETE_EVENT:
			return
		elif response_id == 1:
			return

		self.sync_config ()

	def on_config_checkbutton_toggle (self, button, which):
		self.input_configs[which] = button.get_active ()

	def load_config (self):
		### current configs
		self.configs = {}
		self.configs['show_posters'] = self.gconf.get_bool ('%s/show_posters' % gconf_key)
		self.configs['compatible'] = self.gconf.get_bool ('%s/compatible' % gconf_key)

		### user input configs
		self.input_configs = self.configs.copy ()

	def setup_config_dialog (self):
		dialog = self.config_builder.get_object ("pps_config_dialog")
		dialog.connect ("response", self.on_config_dialog_response)
		dialog.connect ("close", self.on_config_dialog_close)
		dialog.connect ("delete-event", self.on_config_dialog_delete)

		button = self.config_builder.get_object ("pps_config_show_posters")
		button.set_active (self.configs['show_posters'])
		button.connect ("toggled", self.on_config_checkbutton_toggle, 'show_posters')

		button = self.config_builder.get_object ("pps_config_compatible")
		button.set_active (self.configs['compatible'])
		button.connect ("toggled", self.on_config_checkbutton_toggle, 'compatible')

